import gym
from mpl_toolkits.mplot3d import axes3d
from matplotlib import pyplot as plt
import numpy as np

from discrete_q import QLearn
from linear_q import LinQLearn


def episode(env, learner, render=False):
    state = env.reset()
    cum_reward = 0
    done = False
    while not done:
        action = learner.choose_action(state)
        next_state, reward, done, info = env.step(action)
        learner.learn(state, action, reward, next_state)
        if render:
            env.render()
        cum_reward += reward
        state = next_state
    return cum_reward


def train(learner):
    cum_rewards = []
    test_rewards = []
    for i in range(episodes):
        learner.epsilon = learner.epsilon - eps_start / episodes
        cum_reward = episode(env, learner)
        cum_rewards.append(cum_reward)
        if i % (episodes//20) == 0:
            print("episode %d" % i)
            epsilon = learner.epsilon
            learner.epsilon = 0
            test_rewards.append(sum([episode(env, learner) for _ in range(10)]) / 10.)
            learner.epsilon = epsilon
            ax.plot(cum_rewards)
            fig.canvas.draw()
            ax2.plot(test_rewards)
            fig2.canvas.draw()


if __name__ == "__main__":
    env = gym.make('MountainCar-v0')
    env.reset()

    eps_start = 0.9
    alpha = 0.01
    gamma = 0.999
    learner = QLearn(env, eps_start, alpha, gamma)
    # learner = LinQLearn(env, eps_start, alpha, gamma)

    fig,ax = plt.subplots(1,1)
    fig2,ax2 = plt.subplots(1,1)

    episodes = 20000
    # episodes = 500

    train(learner)
    plt.show()

    weights_filename = "weights"+type(learner).__name__+"_"+str(episodes)+".pckl"
    learner.write_q(weights_filename)
    # learner.load_q(weights_filename)

    epsilon = learner.epsilon
    learner.epsilon = 0
    for _ in range(5):
        episode(env, learner, True)
    learner.epsilon = epsilon

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    low = env.observation_space.low
    high = env.observation_space.high
    xs = np.linspace(low[0], high[0], 500)
    ys = np.linspace(low[1], high[1], 500)
    X, Y = np.meshgrid(xs, ys)

    states = np.append(X.reshape(X.shape + (1,)), Y.reshape(Y.shape + (1,)), axis=2)
    states = states.reshape((states.shape[0]*states.shape[1], 2,))
    q = np.array(list(map(lambda x: max([learner.get_q(x, a) for a in learner.actions]), states)))
    Z = -q.reshape(X.shape)

    # plot reward function (multiplied by -1)
    ax.plot_wireframe(X, Y, Z, rstride=10, cstride=10)
    plt.show()

    env.close()
